from __future__ import annotations

import warnings


warnings.warn(  # deprecated in 14.0 - 2024-11-09
    "websockets.legacy is deprecated; "
    "see https://websockets.readthedocs.io/en/stable/howto/upgrade.html "
    "for upgrade instructions",
    DeprecationWarning,
)
